/*
This software is subject to the license described in the license.txt file
included with this software distribution. You may not use this file except in compliance
with this license.

Copyright (c) Dynastream Innovations Inc. 2017
All rights reserved.
*/

#include <stdbool.h>
#include <stdint.h>
#include <string.h>

#include "gfit_defines.h"
#include "gfit_interface.h"

#include "app_error.h"
#include "ant_stack_config.h"
#include "ant_parameters.h"
#include "app_scheduler.h"
#include "softdevice_handler.h"
#include "softdevice_handler_appsh.h"
#include "ble.h"
#include "app_timer.h"
#include "bsp.h"
#include "boards.h"
#include "nrf_nvic.h"
#include "nrf_sdm.h"
#include "nrf_soc.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "sdk_config.h"

// Scheduler settings
#define SCHED_MAX_EVENT_DATA_SIZE       MAX(ANT_STACK_EVT_MSG_BUF_SIZE, BLE_STACK_EVT_MSG_BUF_SIZE)
#define SCHED_QUEUE_SIZE                50

#define BLE_CONN_EVT_LENGTH             15                              /**< 15x1.25ms = 18.75ms (Default = 3*1.25 = 3.75ms) */
#define IS_SRVC_CHANGED_CHARACT_PRESENT 0                               /**< Include or not the service_changed characteristic. if not enabled, the server's database cannot be changed for the lifetime of the device*/

// Proximity pairing settings
// For desk testing with the development kit, the default values for this demo are set to very tight thresholds,
// as this allows experimenting with modules on the same desk (e.g. simulated HRMs).
#define DEMO_SEARCH_THRESHOLD           -31                             /**< Searching threshold. -31 dBm .*/
#define DEMO_PAIR_THRESHOLD             -51                             /**< Pairing threshold. -51dBm .*/

#define DEMO_MFG_SPECIFIC_PAGE          ((uint8_t) 0xE3)                /**< Manufacturer specific page .*/

// FE simulation
#define SIM_BIKE_SPEED_MAX              ((uint16_t) 4)                  /**< Simulated speed max. */
#define SIM_BIKE_CADENCE                ((uint8_t) 90)                  /**< Simulated fixed cadence: 90 RPM. */
#define SIM_BIKE_POWER                  ((uint16_t) 120)                /**< Simulated fixed power: 120 W. */
#define SIM_SPIN_DOWN_TARGET_TIME_MS    ((uint16_t) 10000)              /**< Simulated spin down target time: 10 seconds. */
#define SIM_SPIN_DOWN_TIMEOUT_S         ((uint16_t) 30)                 /**< Simulated spin down timeout: 30 seconds. */

#define MAX_RESISTANCE                  ((uint16_t) 950)                /**< Maximum resistance supported by this equipment (N) */

#define TIMER_TICKS                     APP_TIMER_TICKS(1000)           /**< Timer interval for simulated FE events (1000ms) */
APP_TIMER_DEF(m_timer_id);

// Number of transmissions to send custom event page
#define NUM_TX_EVENT_PAGE               1

#define ENABLED_CONTROLLABLE_FEATURES     { \
                                          .target_resistance = true, \
                                          .target_power = true, \
                                          .target_inclination = true, \
                                          .targeted_distance = true, \
                                          .target_heart_rate = true, \
                                          .reset = true, \
                                          .start_stop_pause = true, \
                                          .bike_simulation = true, \
                                          .spin_down_cal = true, \
                                          .custom = true, \
                                         }

// Enum used to track the state of a simulated spin down calibration.
typedef enum spin_down
{
    spin_down_state_idle,
    spin_down_state_wait_for_target,
    spin_down_state_calibrating
} spin_down_state_t;

static spin_down_state_t m_spin_down_state;
static uint16_t m_spin_down_ticks;

static gfit_state_t m_gfit_state;
static bool m_increase_sim_bike_speed = true;
static bool m_target_power_mode_enabled = false;
static uint16_t sim_bike_speed_m_per_s = 0;
static uint8_t spindown_target_low_m_per_s = 7; // 25.2 kph
static uint8_t spindown_target_high_m_per_s = 9; // 32.4 kph

static gfit_hrp_pairing_mode_t m_pairing_mode = pairing_mode_none; /**< Set to either pairing_mode_proximity to enable RSSI based pairing or pairing_mode_none to disable HR scan. */

bool m_elapsed_time_running = false;
gfit_bike_data_update_t m_bike_trainer_data;                            /**< Stationary Bike Data. */
gfit_simulation_parameters_t m_sim_params;                              /**< Simulation parameters. */
gfit_user_data_t m_user_data;                                           /**< User data (used to calculate simulated resistance. */


static void fe_fields_reset(void);
static void fe_fields_set(void);


/**@brief Function for dispatching a system event to interested modules.
*
* @details This function is called from the System event interrupt handler after a system
*          event has been received.
*
* @param[in]   sys_evt   System stack event.
*/
static void sys_evt_dispatch(uint32_t sys_evt)
{
    gfit_lib_return_t ret = gfit_sys_event_handler(sys_evt);
    APP_ERROR_CHECK_BOOL(GFIT_SUCCESS == ret.gfit_response_code);
}

/**@brief Handle and display scan heart rate events
 * @param[in]   rx_packet   Heart rate packet from G.FIT
 */
static void receive_heart_rate_scan_packet_handler(gfit_hrp_scan_packet_t* rx_packet)
{
    // Packet contains information about heart rate straps visible on scan.
    // If using RSSI based pairing, the scan packets are just for debug purposes,
    // to confirm that HRM straps around G.FIT are being received
    if( protocol_ant == rx_packet->protocol )
    {
        NRF_LOG_DEBUG("Rx Scan Packet (ANT+ %02x-%02x) RSSI: %d\r\n",
                rx_packet->hrm_id[0],
                rx_packet->hrm_id[1],
                rx_packet->rssi);
    }
    else if( protocol_ble == rx_packet->protocol )
    {
        NRF_LOG_DEBUG("Rx Scan Packet (BLE %02x-%02x)  RSSI: %d\r\n",
                rx_packet->hrm_id[0],
                rx_packet->hrm_id[1],
                rx_packet->rssi);
    }
    // Optionally print full 6-byte Device ID at DEBUG level
    NRF_LOG_DEBUG("Device ID:\r\n");
    NRF_LOG_HEXDUMP_DEBUG(rx_packet->hrm_id, GFIT_HRM_ID_LENGTH);
}

/**@brief Handle and display paired heart rate events
 * @param[in]   rx_packet   Heart rate packet from G.FIT
 */
static void receive_heart_rate_paired_packet_handler(gfit_hrp_paired_packet_t* rx_packet)
{
    // Packet contains information about paired heart rate strap
    // Display HR value on console
    NRF_LOG_INFO("Rx Paired HR Packet (%02x-%02x  RSSI %ddBm) HR: %d\r\n",
            rx_packet->hrm_id[0],
            rx_packet->hrm_id[1],
            rx_packet->rssi,
            rx_packet->heart_rate);
    // Optionally print full 6-byte Device ID at DEBUG level
    NRF_LOG_DEBUG("Device ID:\r\n");
    NRF_LOG_HEXDUMP_DEBUG(rx_packet->hrm_id, GFIT_HRM_ID_LENGTH);
}

/**@brief Display a percentage on LEDs
 * Take a percentage and displays it on the LEDs on the dev board.
 *
 * PERCENTAGE        ABCD LEDs
 * -100% and down    ****
 *  -75% to -99%     0***
 *  -50% to -74%     00**
 *  -25% to -49%     000*
 *    0% to -24%     0000
 *    0% to  24%     0000
 *   25% to  49%     *000
 *   50% to  74%     **00
 *   75% to  99%     ***0
 *  100% and up      ****
 *
 * param[in] half_percentage The Percentage to display in half percents.
 */
void display_target_on_leds(int16_t half_percentage)
{
    LEDS_OFF(BSP_LED_0_MASK);
    LEDS_OFF(BSP_LED_1_MASK);
    LEDS_OFF(BSP_LED_2_MASK);
    LEDS_OFF(BSP_LED_3_MASK);

    if(half_percentage < 0)
    {
        if(half_percentage <= -50)
        {
            LEDS_ON(BSP_LED_3_MASK);
        }
        if(half_percentage <= -100)
        {
            LEDS_ON(BSP_LED_2_MASK);
        }
        if(half_percentage <= -150)
        {
            LEDS_ON(BSP_LED_1_MASK);
        }
        if(half_percentage <= -200)
        {
            LEDS_ON(BSP_LED_0_MASK);
        }
    }
    else
    {
        if(half_percentage >= 50)
        {
            LEDS_ON(BSP_LED_0_MASK);
        }
        if(half_percentage >= 100)
        {
            LEDS_ON(BSP_LED_1_MASK);
        }
        if(half_percentage >= 150)
        {
            LEDS_ON(BSP_LED_2_MASK);
        }
        if(half_percentage >= 200)
        {
            LEDS_ON(BSP_LED_3_MASK);
        }
    }
}

/**@brief Function for processing ANT channel events.
 * @param[in] p_ant_evt Pointer to ANT event
 */
static void ant_event_handler(ant_evt_t* p_ant_evt)
{
    // Pass all events managed by G.FIT, as well as device level events (channel field may be zero in these)
    // to the G.FIT ANT event handler
    if ( ( 0 == p_ant_evt->channel) || ( GFIT_ANT_CHANNEL_RESERVED_MIN <= p_ant_evt->channel ) )
    {
        gfit_lib_return_t ret = gfit_ant_event_handler(p_ant_evt);
        APP_ERROR_CHECK_BOOL(GFIT_SUCCESS == ret.gfit_response_code);
    }

    // Handle messages received from ANT+ FE-C display
    if ( ( GFIT_FEC_ANT_CHANNEL_NUMBER_PRIMARY == p_ant_evt->channel ) ||
         ( GFIT_FEC_ANT_CHANNEL_NUMBER_DIVERSIFICATION == p_ant_evt-> channel ) )
    {
        if( EVENT_RX == p_ant_evt->event )
        {
            NRF_LOG_DEBUG("Received message from ANT+ FE-C display.  Payload:\r\n");
            ANT_MESSAGE* p_msg = (ANT_MESSAGE*) p_ant_evt->msg.evt_buffer;
            NRF_LOG_HEXDUMP_DEBUG(p_msg->ANT_MESSAGE_aucPayload, ANT_STANDARD_DATA_PAYLOAD_SIZE);
            // The application can do additional processing of the payload here, e.g.,
            // decoding ANT+ FE-C control pages not yet implemented by G.FIT
            // Note that receiving pages 1, 48, 49, 50, 51, 55 and 70 is already handled by the G.FIT library, and
            // the application should process the corresponding events instead
        }
    }
}

/**@brief Function for processing BLE events.
 * @param[in] p_ble_evt Pointer to BLE event
 */
static void ble_event_handler(ble_evt_t *p_ble_evt)
{
   // Pass all BLE events to the G.FIT BLE event handler
   gfit_lib_return_t ret = gfit_ble_event_handler(p_ble_evt);
   APP_ERROR_CHECK_BOOL(GFIT_SUCCESS == ret.gfit_response_code);
}


/**@brief Custom GFIT event handler
 * Handles a single custom byte event.
 */
static void custom_event_handler(gfit_custom_event_t event)
{
    gfit_lib_return_t ret;
    NRF_LOG_INFO("Received custom event, op code: %d, parameter:\r\n", event.op_code);
    NRF_LOG_HEXDUMP_INFO(event.parameter, GFIT_CUSTOM_EVENT_PARAMETER_SIZE);

    // acknowledge custom commands
    ret = gfit_fep_send_command_update(gfit_command_response_success);
    if(ret.gfit_response_code == GFIT_ERROR_NO_COMMAND_IN_PROGRESS)
    {
        NRF_LOG_WARNING("Tried to send custom command acknowledgment but no command in progress.\r\n");
    }
}

/**@brief Reset Command GFIT event handler
 * Handles the 'Reset' command sent from FTMS Control Point.
 */
static void reset_handler()
{
    gfit_fep_status_update_t update;
    gfit_lib_return_t ret;

    NRF_LOG_INFO("Received reset command over FTMS Control Point. \r\n");

    // Indicate a successfully received reset command.
    ret = gfit_fep_send_command_update(gfit_command_response_success);
    if(ret.gfit_response_code == GFIT_ERROR_NO_COMMAND_IN_PROGRESS)
    {
        NRF_LOG_WARNING("Tried to send reset command acknowledgment but no command in progress. \r\n ")
    }

    // Reset the elapsed time back to 0
    m_bike_trainer_data.elapsed_time_seconds = 0;

    // Reset the target power back to 0
    update.status_type = fe_status_target_power_changed;
    update.target_power_quarter_w = 0;

    // Send the status update for target power
    ret = gfit_fep_send_status_update(update);
    if(ret.gfit_response_code == GFIT_SUCCESS)
    {
        m_target_power_mode_enabled = false;
    }
    else
    {
        NRF_LOG_WARNING("Failed to send target power update. \r\n ");
    }

    // Reset the target resistance level back to 0%
    update.status_type = fe_status_target_resistance_level_changed;
    update.target_resistance_level_half_percent = 0;

    // Send the status update for target resistance level
    ret = gfit_fep_send_status_update(update);
    if(ret.gfit_response_code != GFIT_SUCCESS)
    {
        NRF_LOG_WARNING("Failed to send target resistance update. \r\n ");
    }

    // Send the status update indicating that Reset has been completed
    update.status_type = fe_status_reset;
    ret = gfit_fep_send_status_update(update);

    if(ret.gfit_response_code != GFIT_SUCCESS)
    {
        NRF_LOG_WARNING("Failed to send status update for Reset. \r\n ");
    }
}

/**@brief Stop Command GFIT event handler */
static void stop_handler()
{
    gfit_lib_return_t ret;
    gfit_fep_status_update_t update;

    NRF_LOG_INFO("Received stop command over FTMS Control Point. \r\n");

    if ( m_elapsed_time_running )
    {
        ret = gfit_fep_send_command_update(gfit_command_response_success);
        APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

        // Stop the elapsed time increasing and set back to 0.
        m_elapsed_time_running = false;
        m_bike_trainer_data.elapsed_time_seconds = 0;

        NRF_LOG_INFO("Stopped elapsed time timer.\r\n");

        update.status_type = fe_status_stopped_by_user;
        ret = gfit_fep_send_status_update(update);
        APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
    }
    else
    {
        NRF_LOG_INFO("Unable to stop elapsed time timer, already stopped.\r\n");
        ret = gfit_fep_send_command_update(gfit_command_response_fail);
        APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
    }
}

/**@brief Start/Resume event handler */
static void start_resume_handler()
{
    gfit_lib_return_t ret;
    gfit_fep_status_update_t update;

    NRF_LOG_INFO("Received start command over FTMS Control Point. \r\n");

    if ( !m_elapsed_time_running )
    {
        ret = gfit_fep_send_command_update(gfit_command_response_success);
        APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

        // Start the elapsed time timer.
        m_elapsed_time_running = true;

        NRF_LOG_INFO("Started elapsed time timer.\r\n");

        update.status_type = fe_status_started_or_resumed_by_user;
        ret = gfit_fep_send_status_update(update);
        APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
    }
    else
    {
        NRF_LOG_INFO("Unable to start elapsed time timer, already started.\r\n");

        ret = gfit_fep_send_command_update(gfit_command_response_fail);
        APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
    }
}

/**@brief Pause event handler */
static void pause_handler()
{
    gfit_lib_return_t ret;
    gfit_fep_status_update_t update;

    NRF_LOG_INFO("Received pause command over FTMS Control Point. \r\n");

    if ( m_elapsed_time_running )
    {
        ret = gfit_fep_send_command_update(gfit_command_response_success);
        APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

        // Stop the elapsed time increasing, but do not clear it
        m_elapsed_time_running = false;
        NRF_LOG_INFO("Paused elapsed time timer.\r\n");

        update.status_type = fe_status_paused_by_user;
        ret = gfit_fep_send_status_update(update);
        APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
    }
    else
    {
        NRF_LOG_INFO("Unable to pause elapsed time timer, already stopped.\r\n");
        ret = gfit_fep_send_command_update(gfit_command_response_fail);
        APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
    }
}

static void start_spin_down_handler()
{
    gfit_lib_return_t ret = gfit_fep_spin_down_start(1000 * spindown_target_low_m_per_s, 1000 * spindown_target_high_m_per_s); // Set target speed to about 15-20kph
    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

    NRF_LOG_INFO("Start Spin down Calibration. \r \n");
    NRF_LOG_INFO("Instructing user via display to adjust speed to target: %d m/s - %d m/s \r \n ", spindown_target_low_m_per_s, spindown_target_high_m_per_s);
    m_spin_down_state = spin_down_state_wait_for_target;
    m_spin_down_ticks = 0;
}

static void ignore_spin_down_handler()
{
    // FE had indicated calibration was required, but the user is selecting to ignore that from their phone
    gfit_lib_return_t ret = gfit_fep_spin_down_cancel_request_calibration();
    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

    NRF_LOG_INFO("Canceling calibration request\r\n");
}

static void cancel_spin_down_handler()
{
    // Display has requested to cancel a calibration in progress. Cancel the calibration.
    gfit_lib_return_t ret = gfit_fep_spin_down_cancel(0);
    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

    m_spin_down_state = spin_down_state_idle;
    m_spin_down_ticks = 0;

    NRF_LOG_INFO("Canceling calibration in progress\r\n");
}

static void set_target_resistance_level_handler(int16_t resistance_level_half_percent)
{
    gfit_lib_return_t ret;
    gfit_fep_status_update_t update;

    // Indicate a successfully received command.
    ret = gfit_fep_send_command_update(gfit_command_response_success);
    if(ret.gfit_response_code == GFIT_ERROR_NO_COMMAND_IN_PROGRESS)
    {
        NRF_LOG_WARNING("Tried to send resistance command acknowledgment but no command in progress. \r\n ")
    }

    NRF_LOG_INFO("Received target resistance: " NRF_LOG_FLOAT_MARKER "%%\r \n", NRF_LOG_FLOAT(resistance_level_half_percent/2.0));

    update.status_type = fe_status_target_resistance_level_changed;
    update.target_resistance_level_half_percent = resistance_level_half_percent;

    // Send the response that the target resistance level has been updated
    ret = gfit_fep_send_status_update(update);
    if(ret.gfit_response_code == GFIT_SUCCESS)
    {
        // Trainer is now in basic resistance mode
        m_target_power_mode_enabled = false;

        // Display the target on LEDs (AA dev board only)
        // This simulates the fitness equipment physically setting a resistance on the machine
        display_target_on_leds(resistance_level_half_percent);

        // Update resistance transmitted on broadcast data
        m_bike_trainer_data.resistance_level_half_percent = resistance_level_half_percent;
        m_bike_trainer_data.resistance_level_half_percent_updated = true;
    }
    else
    {
        NRF_LOG_WARNING("Failed to send target resistance update. \r\n ");
    }
}

static void set_target_power_handler(int16_t target_power_quarter_w)
{
    gfit_fep_status_update_t update;
    gfit_lib_return_t ret;

    if (target_power_quarter_w < 0)
    {
        // Negative target power is unsupported.
        // Indicate that the command was unsuccessful.
        ret = gfit_fep_send_command_update(gfit_command_response_fail);
        if(ret.gfit_response_code == GFIT_ERROR_NO_COMMAND_IN_PROGRESS)
        {
            NRF_LOG_WARNING("Tried to send set target power command acknowledgment but no command in progress. \r\n ")
        }

        NRF_LOG_INFO("Received unsupported negative target power: " NRF_LOG_FLOAT_MARKER " Watts\r\n ", NRF_LOG_FLOAT(target_power_quarter_w/4.0));
    }
    else
    {
        // Indicate a successfully received command
        ret = gfit_fep_send_command_update(gfit_command_response_success);
        if(ret.gfit_response_code == GFIT_ERROR_NO_COMMAND_IN_PROGRESS)
        {
            NRF_LOG_WARNING("Tried to send set target power command acknowledgment but no command in progress. \r\n ")
        }

        NRF_LOG_INFO("Received target power: " NRF_LOG_FLOAT_MARKER " Watts\r \n", NRF_LOG_FLOAT(target_power_quarter_w/4.0));

        update.status_type = fe_status_target_power_changed;
        update.target_power_quarter_w = target_power_quarter_w;

        // Send the status indicating target power has been updated
        ret = gfit_fep_send_status_update(update);
        if(ret.gfit_response_code == GFIT_SUCCESS)
        {
            // Trainer is now in target power mode
            m_target_power_mode_enabled = true;
        }
        else
        {
            NRF_LOG_WARNING("Failed to send target power update. \r\n ");
        }
    }
}

static void set_target_inclination_handler(int16_t inclination_hundredth_percent)
{
    gfit_fep_status_update_t update;
    gfit_lib_return_t ret;

    NRF_LOG_INFO("Received target incline: " NRF_LOG_FLOAT_MARKER "%%\r \n", NRF_LOG_FLOAT(inclination_hundredth_percent/100.0));

    ret = gfit_fep_send_command_update(gfit_command_response_success);
    if(ret.gfit_response_code == GFIT_ERROR_NO_COMMAND_IN_PROGRESS)
    {
        NRF_LOG_WARNING("Tried to send target inclination command acknowledgment but no command in progress. \r\n ");
    }

    update.status_type = fe_status_target_incline_changed;
    update.target_inclination_hundredth_percent = inclination_hundredth_percent;

    // Send the status indicating target incline has been updated
    ret = gfit_fep_send_status_update(update);
    if(ret.gfit_response_code == GFIT_SUCCESS)
    {
        // Trainer is now using target inclination
        m_target_power_mode_enabled = false;

        // This example implements a bike, which does not support incline data over broadcast data
        // For equipment types that do support incline (e.g. treadmill), update incline transmitted on broadcast data
    }
    else
    {
        NRF_LOG_WARNING("Failed to send target inclination update. \r\n ");
    }

}

static void set_target_heart_rate_handler(uint8_t heart_rate)
{
    gfit_fep_status_update_t update;
    gfit_lib_return_t ret;

    NRF_LOG_INFO("Received target HR: %d BPM \r \n", heart_rate);

    ret = gfit_fep_send_command_update(gfit_command_response_success);
    if(ret.gfit_response_code == GFIT_ERROR_NO_COMMAND_IN_PROGRESS)
    {
        NRF_LOG_WARNING("Tried to send target heart rate command acknowledgment but no command in progress.\r\n");
    }

    update.status_type = fe_status_target_heart_rate_changed;
    update.target_heart_rate_bpm = heart_rate;

    // Send the status indicating target heart rate has been updated
    ret = gfit_fep_send_status_update(update);
    if(ret.gfit_response_code != GFIT_SUCCESS)
    {
        NRF_LOG_WARNING("Failed to send target heart rate update.\r\n");
    }
}

static void set_targeted_distance_handler(uint32_t targeted_distance_m)
{
    gfit_fep_status_update_t update;
    gfit_lib_return_t ret;

    NRF_LOG_INFO("Received targeted distance: %d m \r \n", targeted_distance_m);

    ret = gfit_fep_send_command_update(gfit_command_response_success);
    if(ret.gfit_response_code == GFIT_ERROR_NO_COMMAND_IN_PROGRESS)
    {
        NRF_LOG_WARNING("Tried to send targeted distance command acknowledgment but no command in progress.\r\n");
    }

    update.status_type = fe_status_targeted_distance_changed;
    update.targeted_distance_m = targeted_distance_m;

    // Send the target update
    ret = gfit_fep_send_status_update(update);
    if(ret.gfit_response_code != GFIT_SUCCESS)
    {
        NRF_LOG_WARNING("Failed to send targeted distance update.\r\n");
    }
}

static void set_simulation_parameters_handler(gfit_simulation_parameters_t params)
{
    gfit_fep_status_update_t update;
    gfit_lib_return_t ret;
    gfit_sim_mode_calculated_settings_t calc;

    NRF_LOG_INFO("Received simulation parameters\r\n");

    if(params.wind_resistance_coefficient_hundredth_kilograms_per_meter_updated)
    {
        NRF_LOG_INFO("Wind Resistance Coefficient: " NRF_LOG_FLOAT_MARKER " kg/m\r\n", NRF_LOG_FLOAT(params.wind_resistance_coefficient_hundredth_kilograms_per_meter/100.0));
        m_sim_params.wind_resistance_coefficient_hundredth_kilograms_per_meter = params.wind_resistance_coefficient_hundredth_kilograms_per_meter;
        m_sim_params.wind_resistance_coefficient_hundredth_kilograms_per_meter_updated = true;
    }
    else if (!m_sim_params.wind_resistance_coefficient_hundredth_kilograms_per_meter_updated)
    {
        m_sim_params.wind_resistance_coefficient_hundredth_kilograms_per_meter = GFIT_DEFAULT_WIND_RESISTANCE_COEFFICIENT_HUNDREDTH_KG_PER_M;
        m_sim_params.wind_resistance_coefficient_hundredth_kilograms_per_meter_updated = true;
    }
    if(params.wind_speed_millimeters_per_second_updated)
    {
        NRF_LOG_INFO("Wind Speed: " NRF_LOG_FLOAT_MARKER " m/s\r\n", NRF_LOG_FLOAT(params.wind_speed_millimeters_per_second/1000.0));
        m_sim_params.wind_speed_millimeters_per_second = params.wind_speed_millimeters_per_second;
        m_sim_params.wind_speed_millimeters_per_second_updated = true;
    }
    else if(!m_sim_params.wind_resistance_coefficient_hundredth_kilograms_per_meter_updated)
    {
        m_sim_params.wind_speed_millimeters_per_second = GFIT_DEFAULT_WIND_SPEED_MMPS;
        m_sim_params.wind_speed_millimeters_per_second_updated = true;
    }
    if(params.drafting_factor_hundredth_updated)
    {
        NRF_LOG_INFO("Drafting factor: " NRF_LOG_FLOAT_MARKER " \r\n", NRF_LOG_FLOAT(params.drafting_factor_hundredth/100.0));
        m_sim_params.drafting_factor_hundredth = params.drafting_factor_hundredth;
        m_sim_params.drafting_factor_hundredth_updated = true;
    }
    else if(!m_sim_params.drafting_factor_hundredth_updated)
    {
        m_sim_params.drafting_factor_hundredth = GFIT_DEFAULT_DRAFTING_FACTOR_HUNDREDTH;
        m_sim_params.drafting_factor_hundredth_updated = true;
    }
    if(params.grade_hundredth_percent_updated)
    {
        NRF_LOG_INFO("Grade: " NRF_LOG_FLOAT_MARKER "%% \r\n", NRF_LOG_FLOAT(params.grade_hundredth_percent/100.0));
        m_sim_params.grade_hundredth_percent = params.grade_hundredth_percent;
        m_sim_params.grade_hundredth_percent_updated = true;
    }
    else if(!m_sim_params.grade_hundredth_percent_updated)
    {
        m_sim_params.grade_hundredth_percent = GFIT_DEFAULT_GRADE_HUNDREDTH_PERCENT;
        m_sim_params.grade_hundredth_percent_updated = true;
    }
    if(params.coefficient_of_rolling_resistance_fifty_micros_updated)
    {
        NRF_LOG_INFO("Coefficient of rolling resistance: %de-6 \r \n", params.coefficient_of_rolling_resistance_fifty_micros*50);
        m_sim_params.coefficient_of_rolling_resistance_fifty_micros = params.coefficient_of_rolling_resistance_fifty_micros;
        m_sim_params.coefficient_of_rolling_resistance_fifty_micros_updated = true;
    }
    else if(!m_sim_params.coefficient_of_rolling_resistance_fifty_micros_updated)
    {
        m_sim_params.coefficient_of_rolling_resistance_fifty_micros = GFIT_DEFAULT_COEFFICIENT_ROLLING_RESISTANCE_FIFTY_MICROS;
        m_sim_params.coefficient_of_rolling_resistance_fifty_micros_updated = true;
    }

    ret = gfit_fep_send_command_update(gfit_command_response_success);
    if(ret.gfit_response_code == GFIT_ERROR_NO_COMMAND_IN_PROGRESS)
    {
        NRF_LOG_WARNING("Tried to send simulation parameters command acknowledgment but no command in progress.\r\n");
    }

    // Calculate simulated resistance
    ret = gfit_fep_calculate_simulated_resistance(m_sim_params, m_user_data, m_bike_trainer_data.instantaneous_speed_millimeters_per_second, &calc);
    if(GFIT_SUCCESS == ret.gfit_response_code)
    {
        NRF_LOG_INFO("Simulated resistance %d mN (%d%%)\r\n", calc.resistance_millinewtons, calc.resistance_millinewtons/(MAX_RESISTANCE * 10));
    }

    update.status_type = fe_status_simulation_parameters_changed;
    update.simulation_parameters = params;

    // Send the response that the simulation parameters have been updated
    ret = gfit_fep_send_status_update(update);
    if(GFIT_SUCCESS == ret.gfit_response_code)
    {
        // Trainer is now in simulation mode
        m_target_power_mode_enabled = false;
    }
    else
    {
        NRF_LOG_WARNING("Failed to send simulation parameters update. \r\n ");
    }
}

static void set_user_data_handler(gfit_user_data_t user_data)
{
    gfit_fep_status_update_t update;
    gfit_lib_return_t ret;

    NRF_LOG_INFO("Received user data configuration\r\n");

    if(user_data.bicycle_wheel_circumference_tenths_millimeters_updated)
    {
        NRF_LOG_INFO("Bicycle wheel circumference: " NRF_LOG_FLOAT_MARKER " mm \r\n", NRF_LOG_FLOAT(user_data.bicycle_wheel_circumference_tenths_millimeters /10.0));
    }
    if(user_data.user_weight_five_grams_updated)
    {
        NRF_LOG_INFO("User weight: %d g\r\n", user_data.user_weight_five_grams * 5);
        m_user_data.user_weight_five_grams = user_data.user_weight_five_grams;
        m_user_data.user_weight_five_grams_updated = true;
    }
    else if(!m_user_data.user_weight_five_grams_updated)
    {
        // Use default value if no value has been provided by display
        m_user_data.user_weight_five_grams = GFIT_DEFAULT_USER_WEIGHT_FIVE_GRAMS;
        m_user_data.user_weight_five_grams_updated = true;
    }
    if(user_data.bicycle_weight_fifty_grams_updated)
    {
        NRF_LOG_INFO("Bicycle weight: %d g\r\n", user_data.bicycle_weight_fifty_grams * 50);
        m_user_data.bicycle_weight_fifty_grams = user_data.bicycle_weight_fifty_grams;
        m_user_data.bicycle_weight_fifty_grams_updated = true;
    }
    else if(!m_user_data.bicycle_weight_fifty_grams_updated)
    {
        // Use default value if no value has been provided by display
        m_user_data.bicycle_weight_fifty_grams = GFIT_DEFAULT_BICYCLE_WEIGHT_FIFTY_GRAMS;
        m_user_data.bicycle_weight_fifty_grams_updated = true;
    }
    if(user_data.gear_ratio_updated)
    {
        NRF_LOG_INFO("Bicycle gear ratio: " NRF_LOG_FLOAT_MARKER " \r\n", NRF_LOG_FLOAT(user_data.gear_ratio * 0.03));
    }

    ret = gfit_fep_send_command_update(gfit_command_response_success);
    if(ret.gfit_response_code == GFIT_ERROR_NO_COMMAND_IN_PROGRESS)
    {
        NRF_LOG_WARNING("Tried to send set user data configuration command acknowledgment but no command in progress.\r\n");
    }

    update.status_type = fe_status_user_data_changed;
    update.user_data = user_data;

    // Send updated user configuration
    ret = gfit_fep_send_status_update(update);
    if(ret.gfit_response_code != GFIT_SUCCESS)
    {
        NRF_LOG_WARNING("Failed to send user data update.\r\n");
    }
}


/**@brief GFIT event handler
 * Handler for events coming back from the G.FIT library.
 */
static void gfit_event_handler(gfit_event_t event)
{
    switch(event.type)
    {
        case gfit_event_receive_heart_rate_scan_packet:
        {
            receive_heart_rate_scan_packet_handler(&event.data.heart_rate_scan_packet);
            break;
        }
        case gfit_event_receive_heart_rate_paired_packet:
        {
            receive_heart_rate_paired_packet_handler(&event.data.heart_rate_paired_packet);
            break;
        }
        case gfit_event_custom:
        {
            custom_event_handler(event.data.custom);
            break;
        }
        case gfit_event_start_spin_down:
        {
            start_spin_down_handler();
            break;
        }
        case gfit_event_ignore_spin_down:
        {
            ignore_spin_down_handler();
            break;
        }
        case gfit_event_cancel_spin_down:
        {
            cancel_spin_down_handler();
            break;
        }
        case gfit_event_set_target_resistance_level:
        {
            set_target_resistance_level_handler(event.data.target_resistance_level_half_percent);
            break;
        }
        case gfit_event_set_target_power:
        {
            set_target_power_handler(event.data.target_power_quarter_w);
            break;
        }
        case gfit_event_set_target_heart_rate:
        {
            set_target_heart_rate_handler(event.data.target_heart_rate_bpm);
            break;
        }
        case gfit_event_set_target_inclination:
        {
            set_target_inclination_handler(event.data.target_inclination_hundredth_percent);
            break;
        }
        case gfit_event_set_targeted_distance:
        {
            set_targeted_distance_handler(event.data.targeted_distance_m);
            break;
        }
        case gfit_event_set_simulation_parameters:
        {
            set_simulation_parameters_handler(event.data.simulation_parameters);
            break;
        }
        case gfit_event_set_user_data:
        {
            set_user_data_handler(event.data.user_data);
            break;
        }
        case gfit_event_reset:
        {
            reset_handler();
            break;
        }
        case gfit_event_start_or_resume:
        {
            start_resume_handler();
            break;
        }
        case gfit_event_stop:
        {
            stop_handler();
            break;
        }
        case gfit_event_pause:
        {
            pause_handler();
            break;
        }
        default:
        {
            break;
        }
    }
}

/**@brief Function for ANT (and BLE) stack initialization.
*
* @details Initializes the SoftDevice and the ANT event interrupt.
*/
static void softdevice_init(void)
{
   uint32_t err_code;

   nrf_clock_lf_cfg_t clock_lf_cfg = NRF_CLOCK_LFCLKSRC;

   // Initialize SoftDevice
   SOFTDEVICE_HANDLER_APPSH_INIT(&clock_lf_cfg, softdevice_evt_schedule);

   // Fetch the start address of the application RAM.
   uint32_t ram_start = 0;
   err_code = softdevice_app_ram_start_get(&ram_start);
   APP_ERROR_CHECK(err_code);

   // BLE stack configuration
   ble_cfg_t ble_cfg;

   // Configure the number of custom UUIDS.
   // - Custom UUID Required for Buttonless DFU Service
   // - Custom UUID Required for GFITCS
   memset(&ble_cfg, 0, sizeof(ble_cfg));
   ble_cfg.common_cfg.vs_uuid_cfg.vs_uuid_count = 2;
   err_code = sd_ble_cfg_set(BLE_COMMON_CFG_VS_UUID, &ble_cfg, ram_start);
   APP_ERROR_CHECK(err_code);

   // Configure the maximum number of connections.
   memset(&ble_cfg, 0, sizeof(ble_cfg));
   ble_cfg.gap_cfg.role_count_cfg.periph_role_count  = GFIT_PERIPHERAL_LINK_COUNT;
   ble_cfg.gap_cfg.role_count_cfg.central_role_count = GFIT_CENTRAL_LINK_COUNT;
   ble_cfg.gap_cfg.role_count_cfg.central_sec_count  = GFIT_CENTRAL_LINK_COUNT;
   err_code = sd_ble_cfg_set(BLE_GAP_CFG_ROLE_COUNT, &ble_cfg, ram_start);
   APP_ERROR_CHECK(err_code);

   // Configure the maximum ATT MTU.
   memset(&ble_cfg, 0x00, sizeof(ble_cfg));
   ble_cfg.conn_cfg.conn_cfg_tag                 = GFIT_CONN_CFG_TAG_CENTRAL;
   ble_cfg.conn_cfg.params.gatt_conn_cfg.att_mtu = NRF_BLE_GATT_MAX_MTU_SIZE;
   err_code = sd_ble_cfg_set(BLE_CONN_CFG_GATT, &ble_cfg, ram_start);
   APP_ERROR_CHECK(err_code);

   // Configure the maximum event length.
   memset(&ble_cfg, 0x00, sizeof(ble_cfg));
   ble_cfg.conn_cfg.conn_cfg_tag                     = GFIT_CONN_CFG_TAG_CENTRAL;
   ble_cfg.conn_cfg.params.gap_conn_cfg.event_length = BLE_CONN_EVT_LENGTH;
   ble_cfg.conn_cfg.params.gap_conn_cfg.conn_count   = GFIT_CENTRAL_LINK_COUNT;
   err_code = sd_ble_cfg_set(BLE_CONN_CFG_GAP, &ble_cfg, ram_start);
   APP_ERROR_CHECK(err_code);

   // Enable BLE stack.
   err_code = softdevice_enable(&ram_start);
   APP_ERROR_CHECK(err_code);

   // Subscribe for BLE events.
   err_code = softdevice_ble_evt_handler_set(ble_event_handler);
   APP_ERROR_CHECK(err_code);

   // Subscribe for ANT events.
   err_code = softdevice_ant_evt_handler_set(ant_event_handler);
   APP_ERROR_CHECK(err_code);

   // Register with the SoftDevice handler module for BLE events.
   err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
   APP_ERROR_CHECK(err_code);

   // Configure ANT stack.
   err_code = ant_stack_static_config();
   APP_ERROR_CHECK(err_code);
}

static void fe_fields_set(void)
{
    gfit_lib_return_t ret;

    // Send all updated bike FE metrics to the G.FIT
    ret = gfit_fep_set_bike_data(m_bike_trainer_data);
    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

    // Once data has been set to G.FIT, reset all update fields
    m_bike_trainer_data.elapsed_time_seconds_updated = false;
    m_bike_trainer_data.total_distance_meters_updated = false;
    m_bike_trainer_data.instantaneous_speed_millimeters_per_second_updated = false;
    m_bike_trainer_data.instantaneous_cadence_half_revolutions_per_min_updated = false;
    m_bike_trainer_data.instantaneous_power_watts_updated = false;
    m_bike_trainer_data.accumulated_power_watts_updated = false;
    m_bike_trainer_data.target_power_limit_updated = false;
    m_bike_trainer_data.resistance_level_half_percent_updated = false;
}

static void fe_common_data_config(void)
{
    gfit_lib_return_t ret;

    gfit_common_data_t demo_common_data = {
                                            .set_manufacturer_name_string = true,
                                            .set_device_name_string = true,
                                            .set_model_number = true,
                                            .device_name_string = "GFITDemoDevice",
                                            .manufacturer_name_string = "DemoManufacturer",
                                            .model_number = 42,
                                          };

    ret = gfit_fep_set_common_data(demo_common_data);
    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
}

static void fe_fields_init(void)
{
    // Configure which FE fields will be supported by this piece of fitness equipment before startup
    m_bike_trainer_data.elapsed_time_seconds_updated = true;
    m_bike_trainer_data.total_distance_meters_updated = true;
    m_bike_trainer_data.instantaneous_speed_millimeters_per_second_updated = true;
    m_bike_trainer_data.instantaneous_cadence_half_revolutions_per_min_updated = true;
    m_bike_trainer_data.instantaneous_power_watts_updated = true;
    m_bike_trainer_data.accumulated_power_watts_updated = true;
    m_bike_trainer_data.target_power_limit_updated = true;
    m_bike_trainer_data.resistance_level_half_percent_updated = true;
}

static void fe_fields_reset(void)
{
    // Reset values
    m_bike_trainer_data.elapsed_time_seconds = 0;
    m_bike_trainer_data.elapsed_time_seconds_updated = true;

    m_bike_trainer_data.total_distance_meters = 0;
    m_bike_trainer_data.total_distance_meters_updated = true;

    m_bike_trainer_data.instantaneous_speed_millimeters_per_second = 0;
    m_bike_trainer_data.instantaneous_speed_millimeters_per_second_updated = true;

    m_bike_trainer_data.instantaneous_cadence_half_revolutions_per_min = 0;
    m_bike_trainer_data.instantaneous_cadence_half_revolutions_per_min_updated = true;

    m_bike_trainer_data.instantaneous_power_watts = 0;
    m_bike_trainer_data.instantaneous_power_watts_updated = true;

    m_bike_trainer_data.accumulated_power_watts = 0;
    m_bike_trainer_data.accumulated_power_watts_updated = true;

    m_bike_trainer_data.target_power_limit = within_target_power_limit_or_target_power_not_set;
    m_bike_trainer_data.target_power_limit_updated = true;

    m_bike_trainer_data.resistance_level_half_percent = 0;
    m_bike_trainer_data.resistance_level_half_percent_updated = true;
}

/**@brief Function for handling BSP events
 * Runs in app context.
 */
static void bsp_event_handler(bsp_event_t evt)
{
    gfit_lib_return_t ret;
    gfit_fep_status_update_t update;

    switch ( evt )
    {
    case BSP_EVENT_KEY_0:
    {
        // Button A = Start
        switch ( m_gfit_state )
        {
        case state_off:
        {
            NRF_LOG_INFO("Set G.FIT state READY\r\n");
            ret = gfit_set_state_ready();
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
            m_gfit_state = ret.current_state;
            fe_fields_reset();
            fe_fields_set();
            break;
        }

        case state_ready:
        {
            NRF_LOG_INFO("Start workout. Set G.FIT state IN USE\r\n");
            ret = gfit_set_state_inuse();
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
            m_gfit_state = ret.current_state;

            // Send status update to indicate user started workout through FE UI
            update.status_type = fe_status_started_or_resumed_by_user;
            ret = gfit_fep_send_status_update(update);
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

            m_elapsed_time_running = true; // Start the elapsed time timer

            break;
        }

        case state_finished:
        {
            NRF_LOG_INFO("Resume workout. Set G.FIT state IN USE\r\n");
            ret = gfit_set_state_inuse();
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
            m_gfit_state = ret.current_state;

            // Send status update to indicate user resumed workout through FE UI
            update.status_type = fe_status_started_or_resumed_by_user;
            ret = gfit_fep_send_status_update(update);
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

            m_elapsed_time_running = true; // Start the elapsed time timer

            break;
        }

        case state_inuse:
        {
            // Starting while IN USE is not a valid operation, so we will
            // reuse the button and use it to send custom data over ANT+ FE-C and BLE
            NRF_LOG_INFO("Send custom event\r\n");

            gfit_custom_event_t custom_event;
            custom_event.op_code = 0xE8; // Op code must be greater than or equal GFIT_MIN_MFG_SPECIFIC_OP_CODE

            // Content and encoding of parameters are manufacturer specific
            custom_event.parameter[0] = 0xCA;
            custom_event.parameter[1] = 0x1A;
            custom_event.parameter[2] = 0xBA;
            custom_event.parameter[3] = 0x2A;
            custom_event.parameter[4] = 0x00;
            ret = gfit_fep_send_custom_event(NUM_TX_EVENT_PAGE, custom_event);
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
            m_gfit_state = ret.current_state;
            break;
        }


        default:
        {
            NRF_LOG_ERROR("Invalid button on this state\r\n");
            break;
        }
        }
        break;
    }

    case BSP_EVENT_KEY_1:
    {
        // Button B = Stop
        switch ( m_gfit_state )
        {
        case state_inuse:
        {
            NRF_LOG_INFO("Pause workout. Set G.FIT state FINISHED\r\n");
            ret = gfit_set_state_finished();
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
            m_gfit_state = ret.current_state;

            m_elapsed_time_running = false; // Pause the elapsed time timer

            // Send status update to indicate user paused workout through FE UI
            update.status_type = fe_status_paused_by_user;
            ret = gfit_fep_send_status_update(update);
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
            break;
        }

        case state_finished:
        {
            NRF_LOG_INFO("End workout. Set G.FIT state READY\r\n");
            // Send status update to indicate user stopped workout through FE UI
            // The status must be sent before switching state to READY, as the BLE peripheral
            // will disconnect on switching to READY state
            update.status_type = fe_status_stopped_by_user;
            ret = gfit_fep_send_status_update(update);
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

            m_bike_trainer_data.elapsed_time_seconds = 0; // Reset the elapsed time timer.
            m_elapsed_time_running = false;

            ret = gfit_set_state_ready();
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
            m_gfit_state = ret.current_state;
            fe_fields_reset();
            fe_fields_set();
            break;
        }

        case state_ready:
        {
            NRF_LOG_INFO("Set G.FIT state OFF\r\n");
            ret = gfit_set_state_off();
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

            m_bike_trainer_data.elapsed_time_seconds = 0; // Reset the elapsed time timer.
            m_elapsed_time_running = false;               // Stop the elapsed time timer.

            m_gfit_state = ret.current_state;
            m_spin_down_state = spin_down_state_idle;
            break;
        }

        case state_off:
        default:
        {
            NRF_LOG_ERROR("Invalid button on this state\r\n");
            break;
        }
        }
        break;
    }

    case BSP_EVENT_KEY_2: // Button C
    {
        // Button C = Disconnect HR / Request/Cancel Calibration
        if ( state_off != m_gfit_state )
        {
            if( pairing_mode_none == m_pairing_mode )
            {
                if ( spin_down_state_idle == m_spin_down_state )
                {
                    NRF_LOG_INFO("FE needs calibration\r\n");
                    ret = gfit_fep_spin_down_request_calibration();
                    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
                }
                else
                {
                    NRF_LOG_INFO("Cancel/Fail calibration in progress\r\n");
                    ret = gfit_fep_spin_down_cancel(0);
                    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
                    m_spin_down_state = spin_down_state_idle;
                }
            }
            else
            {
                NRF_LOG_INFO("Disconnect paired HRM\r\n");
                ret = gfit_hrp_disconnect_device();
                APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

            }
            m_gfit_state = ret.current_state;
        }

        break;
    }

    default:
    {
        // Do nothing
        break;
    }
    }
}


static void spin_down_state_machine(uint16_t sim_bike_speed)
{
    static uint32_t timer_s = 0;
    gfit_lib_return_t ret;

    switch (m_spin_down_state)
    {
    case spin_down_state_idle:
        // No active calibration in progress, do nothing
        break;
    case spin_down_state_wait_for_target:
        m_spin_down_ticks++;
        if(sim_bike_speed >= spindown_target_low_m_per_s && sim_bike_speed <= spindown_target_high_m_per_s)
        {
            NRF_LOG_INFO("Spin down target speed %d m/s has been met.\r\n", sim_bike_speed);
            NRF_LOG_INFO("Notify display that target speed has been reached.\r\n");
            m_spin_down_ticks=0;
            m_spin_down_state = spin_down_state_calibrating;
            m_increase_sim_bike_speed = false;
            ret = gfit_fep_spin_down_set_status(GFIT_INVALID_TEMPERATURE, SIM_SPIN_DOWN_TARGET_TIME_MS, gfit_spin_down_temp_not_applicable, gfit_spin_down_speed_ok);
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
        }
        else if(sim_bike_speed < spindown_target_low_m_per_s)
        {
            // Update status to let display know target speed has not yet been reached
            ret = gfit_fep_spin_down_set_status(GFIT_INVALID_TEMPERATURE, SIM_SPIN_DOWN_TARGET_TIME_MS, gfit_spin_down_temp_not_applicable, gfit_spin_down_speed_too_low);
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
        }

        // If we do not reach the target speed within application specific timeout, cancel the calibration and go back to idle
        if(m_spin_down_ticks >= SIM_SPIN_DOWN_TIMEOUT_S)
        {
            NRF_LOG_INFO("Spin down target speed not reached.  Cancel calibration.\r\n");
            ret = gfit_fep_spin_down_cancel(0);
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
            m_spin_down_ticks = 0;
            m_spin_down_state = spin_down_state_idle;
        }
        break;
    case spin_down_state_calibrating:
        timer_s++;  // Measure calibration time
        // Once the simulated speed reaches zero, simulate a calibration success
        if(sim_bike_speed == 0)
        {
            ret = gfit_fep_spin_down_complete(timer_s * 1000);
            APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
            NRF_LOG_INFO("Spin down Calibration Complete.\r \n")
            NRF_LOG_INFO("Sending calibration time: %d s to display.\r \n", timer_s);
            m_spin_down_state = spin_down_state_idle;
            timer_s = 0;
        }
        else
        {
            // Calibration not completed yet
        }
        break;
    }
}


/**@brief Function for handling timer events
 */
static void timer_event_handler(void * p_context)
{
    switch ( m_gfit_state )
    {
    case state_inuse:
    {
        NRF_LOG_INFO("Simulated Speed:%d m/s \r \n", sim_bike_speed_m_per_s);

        if(sim_bike_speed_m_per_s == 0) m_increase_sim_bike_speed = true;
        if(sim_bike_speed_m_per_s > 15) m_increase_sim_bike_speed = false;

        if(m_increase_sim_bike_speed)
        {
            sim_bike_speed_m_per_s++;
        }
        else
        {
            sim_bike_speed_m_per_s--;
        }

        if (m_target_power_mode_enabled)
        {
            m_bike_trainer_data.target_power_limit_updated = true;
            if (sim_bike_speed_m_per_s > 12)
            {
                m_bike_trainer_data.target_power_limit = undetermined_target_power_limit;
            }
            else if(sim_bike_speed_m_per_s > 8)
            {
                m_bike_trainer_data.target_power_limit = speed_too_high_for_target_power;
            }
            else if(sim_bike_speed_m_per_s > 4)
            {
                m_bike_trainer_data.target_power_limit = within_target_power_limit_or_target_power_not_set;
            }
            else
            {
                m_bike_trainer_data.target_power_limit = speed_too_low_for_target_power;
            }
        }
        else
        {
            m_bike_trainer_data.target_power_limit_updated = true;
            m_bike_trainer_data.target_power_limit = within_target_power_limit_or_target_power_not_set;
        }

        if(m_elapsed_time_running)
        {
            m_bike_trainer_data.elapsed_time_seconds++;
            m_bike_trainer_data.elapsed_time_seconds_updated = true;
        }
        else
        {
            m_bike_trainer_data.elapsed_time_seconds_updated = false;
        }

        m_bike_trainer_data.total_distance_meters += sim_bike_speed_m_per_s;                                // Speed is in m/s, this function gets called every second
        m_bike_trainer_data.total_distance_meters_updated = true;

        m_bike_trainer_data.instantaneous_speed_millimeters_per_second = sim_bike_speed_m_per_s * 1000;     // Convert to mm/s
        m_bike_trainer_data.instantaneous_speed_millimeters_per_second_updated = true;

        m_bike_trainer_data.instantaneous_cadence_half_revolutions_per_min = SIM_BIKE_CADENCE * 2;  // Convert to half revolutions per minute
        m_bike_trainer_data.instantaneous_cadence_half_revolutions_per_min_updated = true;

        m_bike_trainer_data.instantaneous_power_watts = SIM_BIKE_POWER;
        m_bike_trainer_data.instantaneous_power_watts_updated = true;

        m_bike_trainer_data.accumulated_power_watts += SIM_BIKE_POWER;
        m_bike_trainer_data.accumulated_power_watts_updated = true;
        break;
    }

    case state_finished:
    {
        // Continue showing instantaneous fields, do not increase distance or time
        break;
    }

    case state_ready:
    case state_off:
    default:
    {
        fe_fields_reset();   // Reset values while machine not in use
        break;
    }
    }

    spin_down_state_machine(sim_bike_speed_m_per_s);
    fe_fields_set();
}


/**@brief Function for the timer and BSP initialization
 *
 */
static void utils_init(void)
{
    uint32_t err_code;

    err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);

    err_code = bsp_init(BSP_INIT_LED | BSP_INIT_BUTTONS, bsp_event_handler);
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_create(&m_timer_id, APP_TIMER_MODE_REPEATED, timer_event_handler);
    APP_ERROR_CHECK(err_code);
}

/**
 * @brief Main
 */
int main()
{
    uint32_t err_code;
    gfit_lib_return_t ret;
    uint32_t fe_device_id;
    char gfit_lib_version[GFIT_LENGTH_VERSION_STRING];
    gfit_fe_controllable_features_t supported_controllable_features = ENABLED_CONTROLLABLE_FEATURES;

    memset(&m_sim_params, 0, sizeof(gfit_simulation_parameters_t));
    memset(&m_user_data, 0, sizeof(gfit_user_data_t));

    // Initialize RTT logging
    err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);

    // Initialize the soft device
    softdevice_init();

    // Use DC-DC Converter (recommended to optimize power consumption on D52 module)
    err_code = sd_power_dcdc_mode_set(DC_TO_DC_ON);
    APP_ERROR_CHECK(err_code);

    // Initialize the scheduler
    APP_SCHED_INIT(SCHED_MAX_EVENT_DATA_SIZE, SCHED_QUEUE_SIZE);

    // Initialize buttons and timers
    utils_init();

    // Read the UICR ANT ID (to be used as the fe_device_id)
    if(GFIT_ANT_UICR_DEVICE_NUMBER != 0xFFFFFFFF) // 0xFFFFFFFF indicates likely the ANT ID was not set
    {
        fe_device_id = GFIT_ANT_UICR_DEVICE_NUMBER;
    }
    else
    {
        // If ANT ID was not set (e.g. code is running on an SK module, or values stored in UICR have been lost
        // after performing a full erase), set the device ID based on the value in FICR
        // Note that it is highly recommended to use the GFIT_ANT_UICR_DEVICE_NUMBER as the fe_device_id in production.
        // Please refer to the "G.FIT and D52 Premium Module Manufacturing Considerations" for more details on
        // preserving the ANT ID when programming units
        NRF_LOG_WARNING("UICR ANT ID not set, using device ID in FICR\r\n");
        fe_device_id = (uint32_t) NRF_FICR->DEVICEID[0];
    }

    // Initialize G.FIT Library
    // Note: this demo application is designed to be instructional and to alert the user of any issues.
    //       For this reason, the vast majority of error codes are asserted.
    ret = gfit_init(gfit_event_handler, app_error_handler, fe_device_id, GFIT_EVALUATION_LICENSE_KEY);
    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

    m_gfit_state = ret.current_state;

    NRF_LOG_INFO("Initialized G.FIT. Device ID: %lu\r\n", fe_device_id);

    gfit_get_version_string(gfit_lib_version);
    NRF_LOG_INFO("GFIT Library Version Number: %s \n", (uint32_t)gfit_lib_version); // NRF_LOG functions require pointers to be cast as uint32_t

    // Configure G.FIT library
    // Set pairing mode
    ret = gfit_hrp_set_pairing_mode(m_pairing_mode);
    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

    // Only set proximity pairing settings if using proximity pairing mode
    if(m_pairing_mode == pairing_mode_proximity)
    {
        // Configure proximity pairing settings
        ret = gfit_hrp_set_pairing_proximity(DEMO_SEARCH_THRESHOLD, DEMO_PAIR_THRESHOLD);
        APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );
    }

    // Configure FE channels
    // Note that all channels are enabled by default, this configuration is only required if intending to disable any of the features
    gfit_fep_channel_configuration_t config = {.ant_plus_fec_tx = true, .ble_peripheral_tx = true };
    ret = gfit_fep_set_channel_configuration(config);
    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

    // Configure fitness equipment type
    ret = gfit_fep_set_equipment_type(bike, supported_controllable_features);
    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

    // Configure supported FE fields to be transmitted over ANT+ FE-C and BLE FMTS for given fitness equipment type
    fe_fields_init();
    fe_fields_set();

    // Configure common data
    fe_common_data_config();

    // Set the trainers maximum resistance
    ret = gfit_fep_set_max_resistance(MAX_RESISTANCE);
    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

    NRF_LOG_INFO("Configured G.FIT\r\n");

    // Set state to READY
    ret = gfit_set_state_ready();
    APP_ERROR_CHECK_BOOL( GFIT_SUCCESS == ret.gfit_response_code );

    m_gfit_state = ret.current_state;

    NRF_LOG_INFO("Set G.FIT state READY\r\n");

    // Start timer for simulated FE-data
    err_code = app_timer_start(m_timer_id, TIMER_TICKS, NULL);
    APP_ERROR_CHECK(err_code);

    // loop forever
    while (1)
    {
        app_sched_execute();

        NRF_LOG_FLUSH();

        err_code = sd_app_evt_wait();
        APP_ERROR_CHECK(err_code);
    }
}
